/******************************************************
  SCIDYNE Corporation  649 School Street  Pembroke MA 02359  USA  https://scidyne.com
  SCIDYNE_DIO24ARD.cpp - Library Source Code for SCIDYNE DIO24-ARD #100-7692
  Created by Mark Durgin, October 2, 2022   Version 1.0.0

  MIT License https://opensource.org/licenses/MIT

  Copyright 2022 SCIDYNE Corporation or its affiliates.  All Rights Reserved. 

  Permission is hereby granted, free of charge, to any person obtaining a copy of
  this software and associated documentation files (the "Software"), to deal in
  the Software without restriction, including without limitation the rights to
  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
  the Software, and to permit persons to whom the Software is furnished to do so,
  subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*******************************************************/

#include "Arduino.h"
#include "SCIDYNE_DIO24ARD.h"

/******************************************************
 * Call once to initialize the Serial Peripheral Interface and DIO24-ARD
 * - Sets the Arduino I/O pin that will be used exclusively as the DIO24-ARD 
 *   board select and sets to "1" disabling the DIO24-ARD on SPI Bus
 * - Initializes the Arduino SPI to default configuration
 * - Initializes port Data Directions
 */

void SCIDYNE_DIO24ARD::setup( uint16_t pinNumber, uint8_t portA_DDR, uint8_t portB_DDR, uint8_t portC_DDR )
{
  
  selectPin = pinNumber;          // Retain the designated select pin between library calls
  
  pinMode(selectPin, OUTPUT);     // Setup digital output as board select for the DIO24-ARD
  digitalWrite(selectPin, HIGH);  //  and De-Select the DIO24-ARD

  SPI.begin();                    // Activate the Arduino SPI

  pseudoDDR[portA] = portA_DDR;   // Initialize portA Data Direction
  pseudoDDR[portB] = portB_DDR;   // Initialize portB Data Direction
  pseudoDDR[portC] = portC_DDR;   // Initialize portC Data Direction
}


/******************************************************
 * Writes the Data Direction of the designated port
 * where portA = 0,  portB = 1,  portC = 2
 * Bits set to "1" will be outputs, bits set to "0" will be input-only
 */
void SCIDYNE_DIO24ARD::writeDDR ( uint8_t DDR_to_Write, uint8_t DDR_value )
{
  pseudoDDR[DDR_to_Write] = DDR_value;     // Set the Data Direction of designated port 
}


/******************************************************
 * Returns 8-bit Data Direction value of the designated port
 * where portA = 0,  portB = 1,  portC = 2
 * Bits set to "1" are outputs, bits set to "0" are input-only
 */
uint8_t SCIDYNE_DIO24ARD::readDDR( uint8_t DDR_to_Read )
{
  return ( pseudoDDR[DDR_to_Read] );
}


/******************************************************
 * Set a digital bit. Corresponding DDR bit must be "1"
 * Call with bit number, valid range is 0 - 23
 */
void SCIDYNE_DIO24ARD::setBit( uint8_t bitNumber )
{
   uint8_t   port_data[NUMBER_OF_PORTS];        // Array holds the DIO24-ARD ports data
   uint8_t   port_index;                        // Target port based on bitNumber, i.e.; 0-2 
   uint8_t   bit_pos;                           // Target bit i.e.; 0-7
   uint8_t   bit_mask;                          // Holds bit mask
   
   // Read the three ports and store their States
   digitalWrite(selectPin, LOW);                // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_RD);                        // Step 2, Send 1st byte: ReaD command, discard returned byte
   port_data[portA] = SPI.transfer(NULL);       // Step 3, Send 2nd byte: data written is "Don't Care", returns State of PortA
   port_data[portB] = SPI.transfer(NULL);       // Step 4, Send 3rd byte: data written is "Don't Care", returns State of PortB
   port_data[portC] = SPI.transfer(NULL);       // Step 5, Send 4th byte: data written is "Don't Care", returns State of PortC
   digitalWrite(selectPin, HIGH);               // Step 6, De-Select the DIO24-ARD

   port_index = bitNumber / BITS_PER_BYTE;      // Calculate which port, portA=0, portB=1, portC=2
   bit_pos    = bitNumber % BITS_PER_BYTE;      // Calculate which bit
   bit_mask   = 1 << bit_pos;                   // Create a single bit mask

   port_data[port_index] |= bit_mask;           // Set the designated bit

   port_data[portA] &= pseudoDDR[portA];        // Ensure any portA inputs stay inputs
   port_data[portB] &= pseudoDDR[portB];        // Ensure any portB inputs stay inputs
   port_data[portC] &= pseudoDDR[portC];        // Ensure any portC inputs stay inputs
   
   // Write data to the three ports
   digitalWrite(selectPin, LOW);                // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_WR);                        // Step 2, Send 1st byte: WRite Command, discard returned byte
   SPI.transfer(port_data[portA]);              // Step 3, Send 2nd byte: Data to be written to PORTA, discard returned byte
   SPI.transfer(port_data[portB]);              // Step 4, Send 3rd byte: Data to be written to PORTB, discard returned byte
   SPI.transfer(port_data[portC]);              // Step 5, Send 4th byte: Data to be written to PORTC, discard returned byte
   digitalWrite(selectPin, HIGH);               // Step 6, De-Select the DIO24-ARD
}

/******************************************************
 * Clear a digital bit.  Corresponding DDR bit must be "1"
 * Call with bit number, valid range is 0 - 23
 */
void SCIDYNE_DIO24ARD::clrBit( uint8_t bitNumber )
{
   uint8_t   port_data[NUMBER_OF_PORTS];      // Array holds the DIO24-ARD ports data
   uint8_t   port_index;                      // Target port based on bitNumber, i.e.; 0-2 
   uint8_t   bit_pos;                         // Target bit, i.e.; 0-7
   uint8_t   bit_mask;                        // Holds bit mask
   
   // Read the three ports and store their States
   digitalWrite(selectPin, LOW);              // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_RD);                      // Step 2, Send 1st byte: ReaD command, discard returned byte
   port_data[portA] = SPI.transfer(NULL);     // Step 3, Send 2nd byte: data written is "Don't Care", returns State of PORTA
   port_data[portB] = SPI.transfer(NULL);     // Step 4, Send 3rd byte: data written is "Don't Care", returns State of PORTB
   port_data[portC] = SPI.transfer(NULL);     // Step 5, Send 4th byte: data written is "Don't Care", returns State of PORTC
   digitalWrite(selectPin, HIGH);             // Step 6, De-Select the DIO24-ARD

   port_index = bitNumber / BITS_PER_BYTE;    // Calculate which port, portA=0, portB=1, portC=2
   bit_pos    = bitNumber % BITS_PER_BYTE;    // Calculate which bit
   bit_mask   = 1 << bit_pos;                 // Create a single bit mask

   port_data[port_index] &= ~bit_mask;        // Invert the bits, mask bit is now "0"

   port_data[portA] &= pseudoDDR[portA];      // Ensure any portA inputs stay inputs
   port_data[portB] &= pseudoDDR[portB];      // Ensure any portB inputs stay inputs
   port_data[portC] &= pseudoDDR[portC];      // Ensure any portC inputs stay inputs
     
   // Write data to the three ports
   digitalWrite(selectPin, LOW);              // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_WR);                      // Step 2, Send 1st byte: WRite Command, discard returned byte
   SPI.transfer(port_data[portA]);            // Step 3, Send 2nd byte: Data to be written to PORTA, discard returned byte
   SPI.transfer(port_data[portB]);            // Step 4, Send 3rd byte: Data to be written to PORTB, discard returned byte
   SPI.transfer(port_data[portC]);            // Step 5, Send 4th byte: Data to be written to PORTC, discard returned byte
   digitalWrite(selectPin, HIGH);             // Step 6, De-Select the DIO24-ARD
}

/******************************************************
 * Set or Clear a digital bit.  Corresponding DDR bit must be "1"
 * Call with bit number, valid range is 0 - 23 and 
 * desired State, 1 to Set, 0 to Clear
 */
void SCIDYNE_DIO24ARD::bitSetClr(uint8_t bitNumber, bool State)
{
  State ? SCIDYNE_DIO24ARD::setBit( bitNumber ) : SCIDYNE_DIO24ARD::clrBit( bitNumber );
}


/******************************************************
 * Read a digital bit
 * Call with bit number, valid range is 0 - 23
 * Returns State of digital bit
 * NOTE: DIO24-ARD hardware inverts the actual State of the digital channel
 */
bool SCIDYNE_DIO24ARD::readBit( uint8_t bitNumber )
{
   uint8_t   port_data[NUMBER_OF_PORTS];          // Array holds the DIO24-ARD ports data
   uint8_t   port_index;                          // Target port based on bitNumber, i.e.; 0-2 
   uint8_t   bit_pos;                             // Target bit, i.e.; 0-7
   uint8_t   bit_mask;                            // Holds bit mask
   
   // Read the three ports and store their States
   digitalWrite(selectPin, LOW);                  // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_RD);                          // Step 2, Send 1st byte: ReaD command, discard returned byte
   port_data[portA] = SPI.transfer(NULL);         // Step 3, Send 2nd byte: data written is "Don't Care", returns State of PortA
   port_data[portB] = SPI.transfer(NULL);         // Step 4, Send 3rd byte: data written is "Don't Care", returns State of PortB
   port_data[portC] = SPI.transfer(NULL);         // Step 5, Send 4th byte: data written is "Don't Care", returns State of PortC
   digitalWrite(selectPin, HIGH);                 // Step 6, De-Select the DIO24-ARD

   port_index = bitNumber / BITS_PER_BYTE;        // Calculate which port, portA=0, portB=1, portC=2
   bit_pos    = bitNumber % BITS_PER_BYTE;        // Calculate which bit
   bit_mask   = 1 << bit_pos;                     // Create a single bit mask

   return (( port_data[port_index] & bit_mask ) ? true : false);  // Test bit and return Boolean "1" or "0"
}


/******************************************************
 * Write a port
 * Call with port number: 0 = portA, 1 = portB, 2 = portC
 * and 8-bit data to be written. The corresponding Data Direction
 * bits must be set to "1" otherwise a "0" is always written in
 * those positions.
 * NOTE: DIO24-ARD hardware inverts the actual State of the digital channel
 */
void SCIDYNE_DIO24ARD::portWrite(uint8_t port_to_Write, uint8_t port_Byte)
{
   uint8_t   port_data[NUMBER_OF_PORTS];      // Array holds the DIO24-ARD ports data
   
   // Read the three ports and store their States
   digitalWrite(selectPin, LOW);              // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_RD);                      // Step 2, Send 1st byte: ReaD command, discard returned byte
   port_data[portA] = SPI.transfer(NULL);     // Step 3, Send 2nd byte: data written is "Don't Care", returns State of PortA
   port_data[portB] = SPI.transfer(NULL);     // Step 4, Send 3rd byte: data written is "Don't Care", returns State of PortB
   port_data[portC] = SPI.transfer(NULL);     // Step 5, Send 4th byte: data written is "Don't Care", returns State of PortC
   digitalWrite(selectPin, HIGH);             // Step 6, De-Select the DIO24-ARD

   port_data[port_to_Write] = port_Byte;      // Pre-Write the designated port

   port_data[portA] &= pseudoDDR[portA];      // Ensure any portA inputs stay inputs
   port_data[portB] &= pseudoDDR[portB];      // Ensure any portB inputs stay inputs
   port_data[portC] &= pseudoDDR[portC];      // Ensure any portC inputs stay inputs
     
   // Write data to the three ports
   digitalWrite(selectPin, LOW);              // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_WR);                      // Step 2, Send 1st byte: WRite Command, discard returned byte
   SPI.transfer(port_data[portA]);            // Step 3, Send 2nd byte: Data to be written to PORTA, discard returned byte
   SPI.transfer(port_data[portB]);            // Step 4, Send 3rd byte: Data to be written to PORTB, discard returned byte
   SPI.transfer(port_data[portC]);            // Step 5, Send 4th byte: Data to be written to PORTC, discard returned byte
   digitalWrite(selectPin, HIGH);             // Step 6, De-Select the DIO24-ARD
}



/******************************************************
 * Read a port
 * Call with port number: 0 = portA, 1 = portB, 2 = portC
 * Returns 8-bit port data. 
 * NOTE: DIO24-ARD hardware inverts the actual State of the digital channel
 */
uint8_t SCIDYNE_DIO24ARD::portRead(uint8_t port_to_Read)
{
   uint8_t   port_data[NUMBER_OF_PORTS];        // Array holds the DIO24-ARD ports data
   
   // Read the three ports and store their States
   digitalWrite(selectPin, LOW);                // Step 1, Select the DIO24-ARD 
   SPI.transfer(CMD_RD);                        // Step 2, Send 1st byte: ReaD command, discard returned byte
   port_data[portA] = SPI.transfer(NULL);       // Step 3, Send 2nd byte: data written is "Don't Care", returns State of PortA
   port_data[portB] = SPI.transfer(NULL);       // Step 4, Send 3rd byte: data written is "Don't Care", returns State of PortB
   port_data[portC] = SPI.transfer(NULL);       // Step 5, Send 4th byte: data written is "Don't Care", returns State of PortC
   digitalWrite(selectPin, HIGH);               // Step 6, De-Select the DIO24-ARD

   return ( port_data[port_to_Read] );          // Return the desired port data
}
 
